// SPDX-License-Identifier: AGPL-3.0-or-later
use crate::{msg::*, Handle, Inode, server::FsProvider};

use btlib::{bterr, crypto::ConcretePub, IssuedProcRec, Result, BlockPath};
use btmsg::{DeserCallback, Transmitter};

use core::future::{ready, Future, Ready};
use std::{sync::Arc, ops::Deref};
use paste::paste;
use futures::FutureExt;

macro_rules! extractor {
    ($variant:ident) => {
        paste! {
            [<Extract $variant>]
        }
    };
}

macro_rules! reply {
    ($variant:ident) => {
        paste! {
            [<$variant Reply>]
        }
    };
}

macro_rules! extractor_callback {
    ($variant:ident) => {
        paste! {
            struct [<Extract $variant>];

            impl DeserCallback for extractor!($variant) {
                type Arg<'de> = FsReply<'de>;
                type Return = Result<reply!($variant)>;
                type CallFut<'de> = Ready<Self::Return>;
                fn call<'de>(&'de mut self, arg: Self::Arg<'de>) -> Self::CallFut<'de> {
                    let result = if let FsReply::$variant(value) = arg {
                        Ok(value)
                    } else {
                        Err(bterr!("wrong message type sent as reply"))
                    };
                    ready(result)
                }
            }
        }
    };
}

extractor_callback!(Lookup);
extractor_callback!(Create);
extractor_callback!(Open);
extractor_callback!(Write);
extractor_callback!(Link);
extractor_callback!(ReadDir);
extractor_callback!(WriteMeta);

struct AckCallback;

impl DeserCallback for AckCallback {
    type Arg<'de> = FsReply<'de>;
    type Return = Result<()>;
    type CallFut<'de> = Ready<Self::Return>;
    fn call<'de>(&'de mut self, arg: Self::Arg<'de>) -> Self::CallFut<'de> {
        let result = if let FsReply::Ack(_) = arg {
            Ok(())
        } else {
            Err(bterr!("wrong message type sent as reply"))
        };
        ready(result)
    }
}

struct ExtractReadMeta;

impl DeserCallback for ExtractReadMeta {
    type Arg<'de> = FsReply<'de>;
    type Return = Result<ReadMetaReply>;
    type CallFut<'de> = Ready<Self::Return>;
    fn call<'de>(&'de mut self, arg: Self::Arg<'de>) -> Self::CallFut<'de> {
        let result = if let FsReply::ReadMeta(value) = arg {
            Ok(value)
        } else {
            Err(bterr!("wrong message type sent as reply"))
        };
        ready(result)
    }
}

struct ExtractRead<F> {
    callback: Option<F>,
}

impl<F> ExtractRead<F> {
    fn new(callback: F) -> Self {
        Self {
            callback: Some(callback),
        }
    }
}

impl<R, Fut, F> DeserCallback for ExtractRead<F>
where
    F: 'static + Send + FnOnce(ReadReply<'_>) -> Fut,
    Fut: Send + Future<Output = R>,
{
    type Arg<'de> = FsReply<'de> where F: 'de;
    type Return = Result<R>;
    type CallFut<'de> = impl 'de + Send + Future<Output = Self::Return>;
    fn call<'de>(&'de mut self, arg: Self::Arg<'de>) -> Self::CallFut<'de> {
        let callback = self.callback.take().unwrap();
        async move {
            if let FsReply::Read(reply) = arg {
                Ok(callback(reply).await)
            } else {
                Err(bterr!("wrong variant sent in reply to Read message"))
            }
        }
    }
}

pub struct FsClient<T> {
    tx: T,
}

impl<T> FsClient<T> {
    pub fn new(tx: T) -> Self {
        Self { tx }
    }

    pub fn into_inner(self) -> T {
        self.tx
    }

    pub fn get_ref(&self) -> &T {
        &self.tx
    }

    pub fn get_mut(&mut self) -> &mut T {
        &mut self.tx
    }
}

impl<T: Send + Sync + Transmitter> FsProvider for FsClient<T> {
    type LookupFut<'c> = impl 'c  + Send + Future<Output = Result<LookupReply>> where T: 'c;
    fn lookup<'c>(&'c self, _from: &'c Arc<BlockPath>, msg: Lookup<'c>) -> Self::LookupFut<'c> {
        self.tx.call(FsMsg::Lookup(msg), extractor!(Lookup)).map(|e| e?)
    }

    type CreateFut<'c> = impl Send + Future<Output = Result<CreateReply>> where T: 'c;
    fn create<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Create<'c>) -> Self::CreateFut<'c> {
        async move {
            todo!()
        }
    }

    type OpenFut<'c> = impl Send + Future<Output = Result<OpenReply>> where T: 'c;
    fn open<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Open) -> Self::OpenFut<'c> {
        async move {
            todo!()
        }
    }

    type ReadGuard = Vec<u8>;
    type ReadFut<'c> = impl 'c + Send + Future<Output = Result<Self::ReadGuard>> where T: 'c;
    /// Reads from the file specified in the given message.
    /// ### WARNING
    /// The returned guard must be dropped before another method is called on this provider.
    /// Otherwise deadlock _will_ occur.
    fn read<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Read) -> Self::ReadFut<'c> {
        async move {
            todo!()
        }
    }

    type WriteFut<'r> = impl 'r + Send + Future<Output = Result<WriteReply>> where T: 'r;
    fn write<'c>(&'c self, from: &'c Arc<BlockPath>, write: Write<&'c [u8]>) -> Self::WriteFut<'c> {
        async move {
            todo!()
        }
    }

    type FlushFut<'c> = impl 'c + Send + Future<Output = Result<()>> where T: 'c;
    fn flush<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Flush) -> Self::FlushFut<'c> {
        async move {
            todo!()
        }
    }

    type ReadDirFut<'c> = impl 'c + Send + Future<Output = Result<ReadDirReply>> where T: 'c;
    fn read_dir<'c>(&'c self, from: &'c Arc<BlockPath>, msg: ReadDir) -> Self::ReadDirFut<'c> {
        async move {
            todo!()
        }
    }

    type LinkFut<'c> = impl 'c + Send + Future<Output = Result<LinkReply>> where T: 'c;
    fn link<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Link<'c>) -> Self::LinkFut<'c> {
        async move {
            todo!()
        }
    }

    type UnlinkFut<'c> = impl 'c +  Send + Future<Output = Result<()>> where T: 'c;
    fn unlink<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Unlink<'c>) -> Self::UnlinkFut<'c> {
        async move {
            todo!()
        }
    }

    type ReadMetaFut<'c> = impl 'c + Send + Future<Output = Result<ReadMetaReply>> where T: 'c;
    fn read_meta<'c>(&'c self, from: &'c Arc<BlockPath>, msg: ReadMeta) -> Self::ReadMetaFut<'c> {
        async move {
            todo!()
        }
    }

    type WriteMetaFut<'c> = impl 'c + Send + Future<Output = Result<WriteMetaReply>> where T: 'c;
    fn write_meta<'c>(&'c self, from: &'c Arc<BlockPath>, msg: WriteMeta)
        -> Self::WriteMetaFut<'c> {
        async move {
            todo!()
        }
    }

    type AllocateFut<'c> = impl 'c + Send + Future<Output = Result<()>> where T: 'c;
    fn allocate<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Allocate) -> Self::AllocateFut<'c> {
        async move {
            todo!()
        }
    }

    type CloseFut<'c> = impl 'c + Send + Future<Output = Result<()>> where T: 'c;
    fn close<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Close) -> Self::CloseFut<'c> {
        async move {
            todo!()
        }
    }

    type ForgetFut<'c> = impl 'c + Send + Future<Output = Result<()>> where T: 'c;
    fn forget<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Forget) -> Self::ForgetFut<'c> {
        async move {
            todo!()
        }
    }

    type LockFut<'c> = impl 'c + Send + Future<Output = Result<()>> where T: 'c;
    fn lock<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Lock) -> Self::LockFut<'c> {
        async move {
            todo!()
        }
    }

    type UnlockFut<'c> = impl 'c + Send + Future<Output = Result<()>> where T: 'c;
    fn unlock<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Unlock) -> Self::UnlockFut<'c> {
        async move {
            todo!()
        }
    }

    type AddReacapFut<'c> = impl 'c + Send + Future<Output = Result<()>> where T: 'c;
    fn add_readcap<'c>(
        &'c self,
        from: &'c Arc<BlockPath>,
        msg: AddReadcap,
    ) -> Self::AddReacapFut<'c> {
        async move {
            todo!()
        }
    }

    type GrantAccessFut<'c> = impl 'c + Send + Future<Output = Result<()>> where T: 'c;
    fn grant_access<'c>(
        &'c self,
        from: &'c Arc<BlockPath>,
        msg: GrantAccess,
    ) -> Self::GrantAccessFut<'c> {
        async move {
            todo!()
        }
    }
}

impl<T: Transmitter> FsClient<T> {
    /// Lookup up the given name under the given inode.
    pub async fn lookup(&self, parent: Inode, name: &str) -> Result<LookupReply> {
        let msg = FsMsg::Lookup(Lookup { parent, name });
        self.tx.call(msg, extractor!(Lookup)).await?
    }

    pub async fn create(
        &self,
        parent: Inode,
        name: &str,
        flags: Flags,
        mode: u32,
        umask: u32,
    ) -> Result<CreateReply> {
        let msg = FsMsg::Create(Create {
            parent,
            name,
            flags,
            mode,
            umask,
        });
        self.tx.call(msg, extractor!(Create)).await?
    }

    pub async fn open(&self, inode: Inode, flags: Flags) -> Result<OpenReply> {
        let msg = FsMsg::Open(Open { inode, flags });
        self.tx.call(msg, extractor!(Open)).await?
    }

    pub async fn read<R, Fut, F>(
        &self,
        inode: Inode,
        handle: Handle,
        offset: u64,
        size: u64,
        callback: F,
    ) -> Result<R>
    where
        F: 'static + Send + FnOnce(ReadReply<'_>) -> Fut,
        Fut: Send + Future<Output = R>,
    {
        let msg = FsMsg::Read(Read {
            inode,
            handle,
            offset,
            size,
        });
        let callback = ExtractRead::new(callback);
        self.tx.call(msg, callback).await?
    }

    pub async fn write(
        &self,
        inode: Inode,
        handle: Handle,
        offset: u64,
        data: &[u8],
    ) -> Result<WriteReply> {
        let msg = FsMsg::Write(Write {
            inode,
            handle,
            offset,
            data,
        });
        self.tx.call(msg, extractor!(Write)).await?
    }

    pub async fn flush(&self, inode: Inode, handle: Handle) -> Result<()> {
        let msg = FsMsg::Flush(Flush { inode, handle });
        self.tx.call(msg, AckCallback).await?
    }

    pub async fn read_dir(
        &self,
        inode: Inode,
        handle: Handle,
        limit: u32,
        state: u64,
    ) -> Result<ReadDirReply> {
        let msg = FsMsg::ReadDir(ReadDir {
            inode,
            handle,
            limit,
            state,
        });
        self.tx.call(msg, ExtractReadDir).await?
    }

    pub async fn link(&self, inode: Inode, new_parent: Inode, name: &str) -> Result<LinkReply> {
        let msg = FsMsg::Link(Link {
            inode,
            new_parent,
            name,
        });
        self.tx.call(msg, extractor!(Link)).await?
    }

    pub async fn unlink(&self, parent: Inode, name: &str) -> Result<()> {
        let msg = FsMsg::Unlink(Unlink { parent, name });
        self.tx.call(msg, AckCallback).await?
    }

    pub async fn read_meta(&self, inode: Inode, handle: Option<Handle>) -> Result<ReadMetaReply> {
        let msg = FsMsg::ReadMeta(ReadMeta { inode, handle });
        self.tx.call(msg, ExtractReadMeta).await?
    }

    pub async fn write_meta(
        &self,
        inode: Inode,
        handle: Option<Handle>,
        attrs: Attrs,
        attrs_set: AttrsSet,
    ) -> Result<WriteMetaReply> {
        let msg = FsMsg::WriteMeta(WriteMeta {
            inode,
            handle,
            attrs,
            attrs_set,
        });
        self.tx.call(msg, ExtractWriteMeta).await?
    }

    pub async fn allocate(&self, inode: Inode, handle: Handle, size: u64) -> Result<()> {
        let msg = FsMsg::Allocate(Allocate {
            inode,
            handle,
            offset: None,
            size,
        });
        self.tx.call(msg, AckCallback).await?
    }

    pub async fn close(&self, inode: Inode, handle: Handle) -> Result<()> {
        let msg = FsMsg::Close(Close { inode, handle });
        self.tx.call(msg, AckCallback).await?
    }

    pub async fn forget(&self, inode: Inode, count: u64) -> Result<()> {
        let msg = FsMsg::Forget(Forget { inode, count });
        self.tx.call(msg, AckCallback).await?
    }

    pub async fn lock(&self, inode: Inode, handle: Handle, desc: LockDesc) -> Result<()> {
        let msg = FsMsg::Lock(Lock {
            inode,
            handle,
            desc,
        });
        self.tx.call(msg, AckCallback).await?
    }

    pub async fn unlock(&self, inode: Inode, handle: Handle) -> Result<()> {
        let msg = FsMsg::Unlock(Unlock { inode, handle });
        self.tx.call(msg, AckCallback).await?
    }

    pub async fn add_readcap(
        &self,
        inode: Inode,
        handle: Handle,
        pub_creds: ConcretePub,
    ) -> Result<()> {
        let msg = FsMsg::AddReadcap(AddReadcap {
            inode,
            handle,
            pub_creds,
        });
        self.tx.call(msg, AckCallback).await?
    }

    pub async fn grant_access(&self, inode: Inode, record: IssuedProcRec) -> Result<()> {
        let msg = FsMsg::GrantAccess(GrantAccess { inode, record });
        self.tx.call(msg, AckCallback).await?
    }
}

impl<T> AsRef<T> for FsClient<T> {
    fn as_ref(&self) -> &T {
        self.get_ref()
    }
}

impl<T> AsMut<T> for FsClient<T> {
    fn as_mut(&mut self) -> &mut T {
        self.get_mut()
    }
}

impl<T> From<T> for FsClient<T> {
    fn from(value: T) -> Self {
        Self::new(value)
    }
}
